/* ns2t
 * Ns to Tcpdump logfile converter
 *
 * usage: ns2t filename | tcptrace -nS stdin
 */

#include <stdio.h>
#include <unistd.h>
#include <pcap.h>
#include <math.h>
#include <sys/socket.h>
#include <net/bpf.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_compat.h>
#include <netinet/tcp.h>

/* constants */
#define TCPDUMP_MAGIC 0xa1b2c3d4
#define NULL_HDRLEN 4
#define MAX_TCPOPT 40
#define MAX_CAPLEN (sizeof(struct ip) + sizeof(struct tcphdr) + MAX_TCPOPT)
#define MAXLINE 300
#define NSA 3

/* data structures */
struct sackblk {   
  tcp_seq start;          /* start seq no. of sack block */
  tcp_seq end;            /* end seq no. */
};  

struct sackopt {
  unsigned char nop1;
  unsigned char nop2;
  unsigned char kind;
  unsigned char length;
  struct sackblk sackblk[1];
};

/* global variables */
unsigned char nh[NULL_HDRLEN];
unsigned char pkt[MAX_CAPLEN];
static char line[MAXLINE];

/* prototypes */
int fgetline(FILE *fp, char s[], int lim);
void usage(void);

int main(int argc, char **argv)
{
  extern char *optarg;
  extern int optind;
  int ch;
  int oneway = 0; /* one-way mode */
  struct pcap_file_header fh;
  struct pcap_pkthdr ph;
  struct ip *ipb;
  struct tcphdr *tcpb;
  static struct sackopt *sackoptb;
  int linenum = 0;
  int segsize = 0; /* segment size for one-way TCP */
  FILE *fp = stdin;

  while ((ch = getopt(argc, argv, "o")) != -1)
    switch(ch) {
    case 'o':
      oneway = 1;
      break;
    default:
      usage();
      exit(1);
    }

  argc -= optind;
  argv += optind;

  if (argc)
    if ((fp = fopen(*argv, "r")) == NULL) {
      fprintf(stderr, "ns2t: can't open %s\n", *argv);
      exit(1);
    }

  /* build file header */
  fh.magic = TCPDUMP_MAGIC;
  fh.version_major = PCAP_VERSION_MAJOR;
  fh.version_minor = PCAP_VERSION_MINOR;
  fh.thiszone = 0; /* XXX */
  fh.sigfigs = 0; /* XXX */
  fh.snaplen = MAX_CAPLEN + NULL_HDRLEN; /* XXX */
  fh.linktype = DLT_NULL;

  /* write file header */
  fwrite(&fh, sizeof(fh), 1, stdout);

  *((int*) nh) = AF_INET; /* XXX only support AF_INET */
  ipb = (struct ip *) pkt;
  tcpb = (struct tcphdr *) (ipb + 1);
  sackoptb = (struct sackopt *) (tcpb + 1);

  for (;;) {
    /* read the packet info */
    int i;
    char tt;
    double timestamp;
    int junk;
    char type[100];
    int plen;
    char flags[100];
    int iteration;
    int seq;
    int is_ack;
    int is_tcp;
    int rlen;
    u_long src_addr;
    u_long dst_addr;
    u_short src_port;
    u_short dst_port;
    u_short ip_id;
    int ackno;
    int hex_flags;
    int hlen;
	
    /* for SACK option */
    int sa_len;
    int sb[NSA][2];
    int optlen = 0;
    int caplen;
    

    linenum++;

    if (fgetline(fp, line, sizeof(line)) == 0)
      break;

    /* correct NS output line would have 14 or 18 or 24(unofficial) fields: */
    rlen = sscanf(line, "%c %lg %d %d %s %d %s %d %d.%hu %d.%hu %d %hu"
		  " %d %x %d"
		  " %d %d %d %d %d %d %d",
		  &tt,
		  &timestamp,
		  &junk,
		  &junk,
		  type,
		  &plen,
		  flags,
		  &iteration,
		  &src_addr,
		  &src_port,
		  &dst_addr,
		  &dst_port,
		  &seq,
		  &ip_id,     /* 14th field */
		  &ackno,
		  &hex_flags,
		  &hlen,
		  &sa_len,    /* 18th field */
		  &sb[0][0],
		  &sb[0][1],
		  &sb[1][0],
		  &sb[1][1],
		  &sb[2][0],
		  &sb[2][1]   /* 24th field */
		  );

    /*
     * 14 fields : short format for one-way TCP model
     * 18 fields : long format for Full/one-way TCP
     * 24 fields : long format for Full/one-way TCP with sack blocks
     *             (unofficial)
     */
    if (!((oneway && rlen == 14) || rlen == 18 || rlen == 24)) {
      fprintf(stderr,"Bad ns packet header in line %u\n", linenum);
      break;
    }

    is_tcp = strcmp(type, "tcp") == 0;
    is_ack = strcmp(type, "ack") == 0;

    /* if it's not a TCP data segment or ACK, discard and try again */
    if (!is_tcp && !is_ack)
      continue;

    /* capture first seen segment size for one-way TCP
       one-way TCP segment size must be fixed constant. */
    if (oneway && segsize == 0 && is_tcp) {
      segsize = plen - sizeof(struct ip) - sizeof(struct tcphdr);
    }

    ipb->ip_src.s_addr = htonl(src_addr);
    tcpb->th_sport = htons(src_port);
    ipb->ip_dst.s_addr = htonl(dst_addr);
    tcpb->th_dport = htons(dst_port);

    if (rlen == 24 && sa_len) {
      /* build sack option */
      sackoptb->nop1 = TCPOPT_NOP;
      sackoptb->nop2 = TCPOPT_NOP;
      sackoptb->kind = TCPOPT_SACK;
      for (i = 0; i < sa_len; i++) {
	/* fill sack block */
	if (oneway) {
	  sackoptb->sackblk[i].start = htonl((sb[i][0] - 1) * segsize);
	  sackoptb->sackblk[i].end   = htonl((sb[i][1] - 1) * segsize);
	} else {
	  sackoptb->sackblk[i].start = htonl(sb[i][0]);
	  sackoptb->sackblk[i].end   = htonl(sb[i][1]);
	}
      }
      sackoptb->length = 2 + sa_len * 8;
      optlen = 4 + sa_len * 8;

      if (oneway)
	plen += 4; /* ns one-way TCP doesn't care NOP, NOP, SACK, len */
      else
	plen += 2; /* ns FullTcp doesn't care 2 NOPs */
    }

    ipb->ip_len = htons(plen);

    if (oneway) {
      if (is_tcp) {
	tcpb->th_seq = htonl((seq - 1) * segsize);
	tcpb->th_ack = htonl(0);
      } else {
	tcpb->th_seq = htonl(0);
	tcpb->th_ack = htonl(seq * segsize);
	/* fprintf(stderr, "ack is %d\n", seq); */
      }
    } else {
      tcpb->th_seq = htonl(seq);
      tcpb->th_ack = htonl(ackno);
    }

    /* make up a reasonable IPv4 packet header */
    ipb->ip_hl = 5; /* no options, normal length of 20 */
    ipb->ip_v = 4;  /* IPv4 */
    ipb->ip_tos = 0;
    ipb->ip_off = 0;
    ipb->ip_ttl = 64;  /* nice round number */
    ipb->ip_p = 6;     /* TCP */
    ipb->ip_sum = 0;   /* IP checksum, hope it doesn't get checked! */
    ipb->ip_id = htons(ip_id);

    /* is the transport "ECN-Capable"? */
    if (strchr(flags, 'N') != NULL)
      ipb->ip_tos |= IPTOS_ECT;
    
    /* was the "Experienced Congestion" bit set? */
    if (strchr(flags, 'E') != NULL)
      ipb->ip_tos |= IPTOS_CE;

    /* make up a reasonable TCP segment header */
    tcpb->th_off = 5 + optlen / 4; /* now only consider sack option */
    if (rlen == 14 || oneway) {
      tcpb->th_flags = TH_ACK; /* sdo: what about first SYN?? */
      /* fprintf(stderr, "flag is TH_ACK\n"); */
    } else {
      tcpb->th_flags = hex_flags;
    }
    tcpb->th_x2 = 0;
    tcpb->th_sum = 0;
    tcpb->th_urp = 0;
    tcpb->th_win = htons(65535);

    /* x2 *was* reserved, now used for ECN bits */
    if (strchr(flags, 'C') != NULL)
      tcpb->th_x2 |= TH_ECN;
    if (strchr(flags, 'A') != NULL)
      tcpb->th_x2 |= TH_CWR;

    /* build packet header */
    ph.ts.tv_sec = floor(timestamp);
    ph.ts.tv_usec = (timestamp - floor(timestamp)) * 1000000.0;
    caplen = sizeof (struct ip) + sizeof (struct tcphdr) + optlen;
    ph.caplen = NULL_HDRLEN + caplen;

    /*    fprintf(stderr, "ph.caplen = %d\n", ph.caplen); */

    ph.len = NULL_HDRLEN + plen;

    /* write packet header */
    fwrite(&ph, sizeof(ph), 1, stdout);

    /* write NULL link header */
    fwrite(nh, sizeof(nh), 1, stdout);

    /* write packet */
    fwrite(pkt, caplen, 1, stdout);

  }

}

int fgetline(FILE *fp, char s[], int lim)
{
  int c, i;

  i = 0;
  while (--lim > 0 && (c = getc(fp)) != EOF && c != '\n')
    s[i++] = c;
  if (c == '\n')
    s[i++] = c;
  s[i] = '\0';
  return i;
}

void usage(void)
{
  fprintf(stderr,
	  "\n"
	  "usage: ns2t [-o] [filename]\n"
	  "\n"
	  " OPTIONS\n"
	  "\n"
	  "   o one-way TCP model support\n"
	  "\n"
	  " EXAMPLE\n"
	  "\n"
	  "   ns2t sender.tr | tcpdump -r -\n"
	  "\n"
	  "   splits out.tr | ns2t | tcptrace -nS stdin\n"
	  );
}
